
(*
 *	Copyright (c) 2002 by Laboratoire Spcification et Vrification (LSV),
 *	CNRS UMR 8643 & ENS Cachan.
 *	Written by Jean Goubault-Larrecq.  Not derived from licensed software.
 *
 *	Permission is granted to anyone to use this software for any
 *	purpose on any computer system, and to redistribute it freely,
 *	subject to the following restrictions:
 *
 *	1. Neither the author nor its employer is responsible for the consequences of use of
 *		this software, no matter how awful, even if they arise
 *		from defects in it.
 *
 *	2. The origin of this software must not be misrepresented, either
 *		by explicit claim or by omission.
 *
 *	3. Altered versions must be plainly marked as such, and must not
 *		be misrepresented as being the original software.
 *
 *	4. This software is restricted to non-commercial use only.  Commercial
 *		use is subject to a specific license, obtainable from LSV.
 *		This was developed as part of the EVA project; as such, access
 *		to this software by partners of the RNTL EVA Project (Trusted
 *		Logic S.A., Verimag) is defined by the intellectual property
 *		agreement signed as part of this project.
 *)

open "conclo_h"

fun cc () =
    let (* First, we assume a so-called universe: this is a finite
	 set of terms, closed under subterms. *)

	(* The parents of a color c are all terms f (..., t, ...) in the universe
	 with t of color c. *)
	val parents = ref ({} : eva_color -m> eva_term set ref (* the set of parents of the color *)
			     )

	(* color maps each term in the universe to its color,
	 i.e., a unique term in its equivalence class. *)
	val color = ref ({} : eva_term -m> eva_color)

	(* invcolor is the inverse map to color: it yields the set of
	 all terms of a given color. *)
	val invcolor = ref ({} : eva_color -m> eva_term set ref)

	(* sigtable maps f (c1, ..., cn) where each ci is a color, to some term in the universe
	 of the color of f (c1, ..., cn).
	 This is provided there is t1 of color c1, ..., tn of color cn,
	 such that f (t1, ..., tn) is in the universe. *)
	val sigtable = ref ({} : string * eva_color list -m> eva_term)

	(* pending holds a list of equations u=v that need to be taken care of. *)
	val pending = ref (nil : (eva_term * eva_term) list)

	fun update_parents (cl, c) =
	    iterate
	      let val prts = ?(!parents) cu
	      in
		  prts := !prts U {c}
	      end
	    | cu in list cl
	    end

	fun find t =
	    let val clr = !color
	    in
		if t inset clr
		    then ?clr t
		else case t of
			 T_VAR _ => (color := clr ++ {t => t};
				     invcolor := !invcolor ++ {t => ref {t}};
				     parents := !parents ++ {t => ref {}};
				     t)
		       | T_APP (f, l) => let val cl = [find u | u in list l]
					     val clr = !color
					     val sg = (f, cl)
					     val sigt = !sigtable
					 in
					     if sg inset sigt (* the new term in fact already has a color: *)
						 then let val tt = ?sigt sg
							  val c = ?clr tt
							  val invc = ?(!invcolor) c
						      in
							  color := clr ++ {t => c};
							  invc := !invc U {t};
							  update_parents (cl, c);
							  c
						      end
					     else (* otherwise create a new color: this will be t itself. *)
						 (parents := !parents ++ {t => ref {}};
						  update_parents (cl, t);
						  color := clr ++ {t => t};
						  invcolor := !invcolor ++ {t => ref {t}};
						  sigtable := sigt ++ {sg => t};
						  t)
					 end
	    end

	fun killsigs paru = (* kills every signature entry that is congruent to some
			     term in paru.  Returns a list of equalities (parent, t)
			     that should still hold.
			     congruent is the relation defined so that
			     f (s1,...,sn) congruent f (t1,...,tn) iff
			     find si=find ti for each i. *)
	    let val sigt = !sigtable
		val vanishing = [let val cargs = [find t | t in list args]
				     val sg = (f, cargs)
				 in
				     sigtable := {sg} <-| !sigtable;
				     (parent, ?sigt sg)
				 end
				| parent as T_APP (f, args) in set paru]
	    in
		vanishing
	    end

	fun reinstallsigs eqns = (* recomputes signature entries corresponding
				  to each application term in paru. *)
	    iterate
	      let val cargs = [find t | t in list args]
		  val sg = (f, cargs)
	      in
		  if sg inset !sigtable
		      then pending := (?(!sigtable) sg, tt) :: !pending
		  else sigtable := !sigtable ++ {sg => tt}
	      end
	    | (parent as T_APP (f, args), tt) in list eqns
	    end

	fun merge_colors (cu, cv) = (* rewrite color cu to cv in color and invcolor tables. *)
	    let val invclr = !invcolor
		val us = !(?invclr cu)
	    in
		color := !color ++ {u => cv
				   | u in set us};
		invcolor := ({cu} <-| invclr);
		let val vsr = ?invclr cv
		in
		    vsr := !vsr U us
		end
	    end

	fun merge (cu, paru, cv) = (* rewrite color cu to cv everywhere;
				    paru is the set of application terms in
				    the universe that have some argument of
				    color cu. *)
	    let val eqns = killsigs paru
	    in
		merge_colors (cu, cv);
		reinstallsigs eqns
	    end

	fun do_union (u, v) =
	    let val cu = find u
		val cv = find v
	    in
		if cu=cv
		    then ()
		else let val paru = !(?(!parents) cu)
			 val parv = !(?(!parents) cv)
		     in
			 if card paru<card parv
			     then merge (cu, paru, cv)
			 else merge (cv, parv, cu)
		     end
	    end

	fun process () =
	    case !pending of
		eqn::rest => (pending := rest;
			      do_union eqn;
			      process ())
	      | _ => ()

	fun equate eqn = (do_union eqn; process ())

	fun class c =
	    let val invc = !invcolor
	    in
		if c inset invc
		    then !(?invc c)
		else raise NotColor
	    end

	fun select (fs, c) =
	    {t
	    | t as T_APP (f, _) in set class c
	      such that f inset fs}


	fun add_rewrite (u, v, vars) = (* adds all equations u.sigma = v.sigma,
					where sigma only binds variables in vars,
					and u.sigma is in the current universe.
					    Assumes FV(v) subseteq FV(u), i.e.,
					that u -> v is a proper rewrite rule. *)
	    let exception MatchFailEvt
		val sr = ref ({} : eva_subst)
		fun match1 (s as T_VAR x, t) =
		    if x inset !sr
			then if ?(!sr) x=find t
				 then ()
			     else raise MatchFailEvt
		    else if x inset vars
			then sr := !sr ++ {x => find t}
		    else if find s=find t
			then ()
		    else raise MatchFailEvt
		  | match1 (T_APP (f, pl), T_APP (g, tl)) =
		    if f=g
			then
			    iterate
			      match1 (p, t)
			    || p in list pl and t in list tl
			    end
		    else raise MatchFailEvt
		  | match1 _ = raise MatchFailEvt

		fun match pat_t = ((sr := {};
				    (match1 pat_t; SOME (!sr))
				    handle MatchFailEvt => NONE)
				       before sr:={})
		fun subst (sigma, T_VAR x) =
		    if x inset sigma
			then ?sigma x
		    else if x inset vars
			then raise NotRewrite x
		    else T_VAR x
		  | subst (sigma, T_APP (f, l)) =
		    T_APP (f, [subst (sigma, t) | t in list l])
		    
		val sigmas = {(t, sigma)
			     | t in set !color
			       val SOME sigma = match (u, t)}
	    in
		pending := [(t, subst (sigma, v))
			   | (t, sigma) in set sigmas] @ !pending;
		process ()
	    end
    in
	|[ find = find,
	   equate = equate,
	   class = class,
	   select = select,
	   add_rewrite = add_rewrite
	   ]|
    end;

