
(*
 *	Copyright (c) 2002 by Laboratoire Spcification et Vrification (LSV),
 *	CNRS UMR 8643 & ENS Cachan.
 *	Written by Jean Goubault-Larrecq.  Not derived from licensed software.
 *
 *	Permission is granted to anyone to use this software for any
 *	purpose on any computer system, and to redistribute it freely,
 *	subject to the following restrictions:
 *
 *	1. Neither the author nor its employer is responsible for the consequences of use of
 *		this software, no matter how awful, even if they arise
 *		from defects in it.
 *
 *	2. The origin of this software must not be misrepresented, either
 *		by explicit claim or by omission.
 *
 *	3. Altered versions must be plainly marked as such, and must not
 *		be misrepresented as being the original software.
 *
 *	4. This software is restricted to non-commercial use only.  Commercial
 *		use is subject to a specific license, obtainable from LSV.
 *		This was developed as part of the EVA project; as such, access
 *		to this software by partners of the RNTL EVA Project (Trusted
 *		Logic S.A., Verimag) is defined by the intellectual property
 *		agreement signed as part of this project.
 *)

open "lispin_h"

fun trans_warn (msg,
		|[first_line, first_column, last_line, last_column, in_file]| : lineinfo) =
    (#put stderr "evapp: ";
     if in_file<>""
	 then (#put stderr in_file;
	       #put stderr ", line";
	       if first_line<>last_line then #put stderr "s" else ();
	       #put stderr " ";
               print stderr (pack first_line);
	       if first_column<>0 then (#put stderr "("; print stderr (pack first_column);
					#put stderr ")")
		   else ();
	       if first_line<>last_line orelse first_column<>last_column
		   then (#put stderr "-";
			 print stderr (pack last_line);
			 if (last_column<>0) then (#put stderr "("; print stderr (pack last_column);
						   #put stderr ")")
			 else ())
	       else ())
     else ();
     #put stderr ": ";
     #put stderr msg;
     #put stderr "\n";
     #flush stderr ());

fun trans_fatal msg_linfo = (trans_warn msg_linfo;
			     quit 10);

val trans_err_limit = ref 20;

fun trans_error msg_linfo =
    (trans_warn msg_linfo;
     dec trans_err_limit;
     if !trans_err_limit<=0
	 then quit 10
     else ());

exception FailEvt of string;

fun format_error (s, APP (f, _)) =
    raise FailEvt ("format error (" ^ s ^ "), got " ^ f)
  | format_error (s, INT n) =
    let val f as |[put, convert, ...]| = outstring "format error ("
    in
	put s;
	put "), got int ";
	print f (pack n);
	raise FailEvt (convert ())
    end
  | format_error (s, VAR f) =
    raise FailEvt ("format error (" ^ s ^ "), got variable " ^ f)
  | format_error (s, LINE (t, linfo)) =
    (format_error (s, t) handle FailEvt msg => trans_fatal (msg, linfo));

val |[put=oput, flush=oflush, ...]| = stdout;

fun pp_type_dom nil = ()
  | pp_type_dom l =
    (let val sep = ref ""
     in
	 iterate
	   (oput (!sep);
	    sep := ", ";
	    oput x)
	 | VAR x in list l
	 end
     end);

fun pp_function_type (args, VAR x) =
    (oput "(";
     pp_type_dom args;
     oput ") : ";
     oput x)
  | pp_function_type (_, t) = format_error ("function result type", t);

fun pp_eva_type (LINE (t, linfo)) =
    (pp_eva_type t handle FailEvt msg => trans_fatal (msg, linfo))
  | pp_eva_type (APP ("function", res::args)) =
    pp_function_type (args, res)
  | pp_eva_type (APP ("one-way-function", res::args)) =
    (pp_function_type (args, res);
     oput " hash")
  | pp_eva_type (APP ("secret-function", res::args)) =
    (pp_function_type (args, res);
     oput " secret")
  | pp_eva_type (APP ("secret-one-way-function", res::args)) =
    (pp_function_type (args, res);
     oput " hash secret")
  | pp_eva_type (VAR x) = (oput " : "; oput x)
  | pp_eva_type t = format_error ("function type", t);

fun pp_type1 (LINE (t, linfo)) =
    (pp_type1 t handle FailEvt msg => trans_fatal (msg, linfo))
  | pp_type1 (APP ("type", [VAR x, typ])) =
    (oput x;
     pp_eva_type typ)
  | pp_type1 t = format_error ("type", t);

fun pp_type t = (oput "  "; pp_type1 t; oput "\n");

fun pp_eva_term (LINE (t, linfo)) =
    (pp_eva_term t handle FailEvt msg => trans_fatal (msg, linfo))
  | pp_eva_term (VAR x) = oput x
  | pp_eva_term (APP ("crypt", [a, m, k])) =
    (oput "{";
     pp_eva_term m;
     oput "}_(";
     pp_eva_term k;
     oput ")";
     pp_eva_term_alg a)
  | pp_eva_term (APP ("sign", [a, m, k])) =
    (oput "[";
     pp_eva_term m;
     oput "]_(";
     pp_eva_term k;
     oput ")";
     pp_eva_term_alg a)
  | pp_eva_term (APP ("tuple", l)) = pp_eva_terml l
  | pp_eva_term (APP ("d", [t])) = pp_eva_term t
  | pp_eva_term (APP ("p", [t])) = pp_eva_term t
  | pp_eva_term (APP ("a", [t])) = pp_eva_term t
  | pp_eva_term (APP ("sa", [t])) = pp_eva_term t
  | pp_eva_term (APP ("aa", [t])) = pp_eva_term t
  | pp_eva_term (APP ("vanilla", nil)) = oput "vanilla"
  | pp_eva_term (APP ("apply", VAR f::l)) = (oput f; pp_eva_terml l)
  | pp_eva_term (APP ("hash-apply", l)) = pp_eva_term (APP ("apply", l))
  | pp_eva_term (APP ("secret-apply", VAR f::l)) = (oput f; pp_eva_terml l)
  | pp_eva_term (APP ("hash-secret-apply", l)) = pp_eva_term (APP ("apply", l))
  | pp_eva_term (APP ("apply-pubk", alg::VAR s::l)) = pp_eva_term_pubk (s, alg, l)
  | pp_eva_term (APP ("hash-apply-pubk", alg::VAR s::l)) = pp_eva_term_pubk (s, alg, l)
  | pp_eva_term (APP ("apply-privk", alg::VAR s::l)) = pp_eva_term_privk (s, alg, l)
  | pp_eva_term (APP ("hash-apply-privk", alg::VAR s::l)) = pp_eva_term_privk (s, alg, l)
  | pp_eva_term (APP ("lambda-pubk", alg::l)) = (oput "lambda-pubk"; pp_eva_term_alg alg; pp_eva_terml l)
  | pp_eva_term (APP ("lambda-privk", alg::l)) = (oput "lambda-privk"; pp_eva_term_alg alg; pp_eva_terml l)
  | pp_eva_term (APP ("let", [APP ("exact", [s]), t])) =
    (pp_eva_term s;
     oput " = ";
     pp_eva_term t)
  | pp_eva_term (APP ("let", [pat, t])) =
    (pp_eva_pattern pat;
     oput " := ";
     pp_eva_term t)
  | pp_eva_term (APP ("new", [pat])) =
    (oput "new ";
     pp_eva_pattern pat)
  | pp_eva_term (APP ("recv", [pat])) =
    (oput "recv ";
     pp_eva_pattern pat)
  | pp_eva_term (APP ("skip", nil)) =
    oput "skip"
  | pp_eva_term (APP ("send", [t])) =
    (oput "send ";
     pp_eva_term t)
  | pp_eva_term (APP ("located", [VAR s, VAR p, t])) =
    (oput "("; pp_eva_term t; oput ")"; oput "@"; oput s; oput p)
  | pp_eva_term (t as APP (f, l)) =
    if f inset {"prop", "at", "owns", "secret", "not", "E", "A", "N", "F", "G", "P", "Q", "and", "or",
		 "U", "W", "S", "B", "implies"}
	then pp_eva_imp t
    else (oput f; oput "$"; pp_eva_terml l)
  | pp_eva_term t = format_error ("term", t)
and pp_eva_atom (VAR x) = oput x
  | pp_eva_atom (APP ("prop", VAR p::l)) =
    (oput p; oput " "; pp_eva_terml l)
  | pp_eva_atom (APP ("at", [VAR s, VAR p, VAR lab])) =
    (oput s; oput p; oput "@"; oput lab)
  | pp_eva_atom (APP ("owns", [VAR s, VAR p, t])) =
    (oput s; oput p; oput "("; pp_eva_term t; oput ")")
  | pp_eva_atom (APP ("secret", [t])) =
    (oput "secret("; pp_eva_term t; oput ")")
  | pp_eva_atom (APP ("not", [F])) = (oput "!"; pp_eva_atom F)
  | pp_eva_atom (APP ("E", [F])) = (oput "*E "; pp_eva_atom F)
  | pp_eva_atom (APP ("A", [F])) = (oput "*A "; pp_eva_atom F)
  | pp_eva_atom (APP ("N", [F])) = (oput "*N "; pp_eva_atom F)
  | pp_eva_atom (APP ("F", [F])) = (oput "*F "; pp_eva_atom F)
  | pp_eva_atom (APP ("G", [F])) = (oput "*G "; pp_eva_atom F)
  | pp_eva_atom (APP ("P", [F])) = (oput "*P "; pp_eva_atom F)
  | pp_eva_atom (APP ("Q", [F])) = (oput "*Q "; pp_eva_atom F)
  | pp_eva_atom (APP ("and", nil)) = oput "true"
  | pp_eva_atom (APP ("or", nil)) = oput "false"
  | pp_eva_atom F = (oput "("; pp_eva_imp F; oput ")")
and pp_eva_and (APP ("and", l)) =
    pp_eva_and1 l
  | pp_eva_and F = pp_eva_atom F
and pp_eva_and1 (F::G::rest) =
    (pp_eva_atom F; oput " && ";
     if null rest
	 then pp_eva_and G
     else pp_eva_and1 (G::rest))
  | pp_eva_and1 [F] = pp_eva_atom F
  | pp_eva_and1 nil = oput "true"
and pp_eva_or (APP ("or", l)) =
    pp_eva_or1 l
  | pp_eva_or (APP ("U", [F, G])) = (pp_eva_and F; oput " *U "; pp_eva_or G)
  | pp_eva_or (APP ("W", [F, G])) = (pp_eva_and F; oput " *W "; pp_eva_or G)
  | pp_eva_or (APP ("S", [F, G])) = (pp_eva_and F; oput " *S "; pp_eva_or G)
  | pp_eva_or (APP ("B", [F, G])) = (pp_eva_and F; oput " *B "; pp_eva_or G)
  | pp_eva_or F = pp_eva_and F
and pp_eva_or1 (F::G::rest) = 
    (pp_eva_and F; oput " || ";
     if null rest
	 then pp_eva_or G
     else pp_eva_or1 (G::rest))
  | pp_eva_or1 [F] = pp_eva_and F
  | pp_eva_or1 nil = oput "false"
and pp_eva_imp (APP ("implies", [F, G])) =
    (pp_eva_or F; oput " => "; pp_eva_imp G)
  | pp_eva_imp F = pp_eva_or F
and pp_eva_pattern (LINE (t, linfo)) =
    (pp_eva_pattern t handle FailEvt msg => trans_fatal (msg, linfo))
  | pp_eva_pattern (VAR x) = oput x
  | pp_eva_pattern  (APP ("crypt", [a, m, k])) =
    (oput "{";
     pp_eva_pattern m;
     oput "}_(";
     pp_eva_term k;
     oput ")";
     pp_eva_term_alg a)
  | pp_eva_pattern (APP ("sign", [a, m, k])) =
    (oput "[";
     pp_eva_pattern m;
     oput "]_(";
     pp_eva_term k;
     oput ")";
     pp_eva_term_alg a)
  | pp_eva_pattern (APP ("exact", [t])) =
    (oput "<";
     pp_eva_term t;
     oput ">")
  | pp_eva_pattern (APP ("tuple", l)) = pp_eva_patternl l
  | pp_eva_pattern (APP ("d", [t])) = pp_eva_pattern t
  | pp_eva_pattern (APP ("p", [t])) = pp_eva_pattern t
  | pp_eva_pattern (APP ("a", [t])) = pp_eva_pattern t
  | pp_eva_pattern (APP ("sa", [t])) = pp_eva_pattern t
  | pp_eva_pattern (APP ("aa", [t])) = pp_eva_pattern t
  | pp_eva_pattern (APP ("vanilla", nil)) = oput "vanilla"
  | pp_eva_pattern (APP ("apply", VAR f::l)) = (oput f; pp_eva_patternl l)
  | pp_eva_pattern (APP ("hash-apply", l)) = pp_eva_pattern (APP ("apply", l))
  | pp_eva_pattern (APP ("secret-apply", l)) = pp_eva_pattern (APP ("apply", l))
  | pp_eva_pattern (APP ("hash-secret-apply", l)) = pp_eva_pattern (APP ("apply", l))
  | pp_eva_pattern (APP ("apply-pubk", alg::VAR s::l)) = pp_eva_term_pubk (s, alg, l)
  | pp_eva_pattern (APP ("hash-apply-pubk", alg::VAR s::l)) = pp_eva_term_pubk (s, alg, l)
  | pp_eva_pattern (APP ("apply-privk", alg::VAR s::l)) = pp_eva_term_privk (s, alg, l)
  | pp_eva_pattern (APP ("hash-apply-privk", alg::VAR s::l)) = pp_eva_term_privk (s, alg, l)
  | pp_eva_pattern (APP (f, l)) = (oput f; oput "$"; pp_eva_patternl l)
  | pp_eva_pattern t = format_error ("pattern", t)
and pp_eva_terml nil = oput "()"
  | pp_eva_terml (a::l) = (oput "("; pp_eva_term a;
			   iterate
			     (oput ","; pp_eva_term t)
			   | t in list l
			   end;
			   oput ")")
and pp_eva_patternl nil = oput "()"
  | pp_eva_patternl (a::l) = (oput "("; pp_eva_pattern a;
			      iterate
				(oput ","; pp_eva_pattern t)
			      | t in list l
			      end;
			      oput ")")
and pp_eva_term_pubk (name, alg, l) =
    (oput "apply-pubk"; pp_eva_term_alg alg; pp_eva_terml (VAR name::l))
and pp_eva_term_privk (name, alg, l) =
  (oput "apply-privk"; pp_eva_term_alg alg; pp_eva_terml (VAR name::l))
and pp_eva_term_alg (APP ("sa", [APP ("vanilla", nil)])) = ()
  | pp_eva_term_alg a = (oput "^("; pp_eva_term a; oput ")");

fun pp_alias (LINE (t, linfo)) =
    (pp_alias t handle FailEvt msg => trans_fatal (msg, linfo))
  | pp_alias (APP ("alias", [VAR x, t])) =
    (oput "  alias ";
     oput x;
     oput " = ";
     pp_eva_term t;
     oput "\n")
  | pp_alias t = format_error ("alias", t);

fun pp_axiom (LINE (t, linfo)) =
    (pp_axiom t handle FailEvt msg => trans_fatal (msg, linfo))
  | pp_axiom (APP ("axiom", s::t::typ::l)) =
    (oput "  axiom ";
     pp_eva_term s;
     oput " = ";
     pp_eva_term t;
     pp_eva_type typ;
     (case l of
	  t::l' => (oput " [";
		    pp_type1 t;
		    iterate
		      (oput ", ";
		       pp_type1 t)
		    | t in list l'
		    end;
		    oput "]")
	| _ => ()
	  );
     oput "\n")
  | pp_axiom t = format_error ("axiom", t);

fun pp_transition (LINE (t, linfo)) =
    (pp_transition t handle FailEvt msg => trans_fatal (msg, linfo))
  | pp_transition (APP ("trans", [VAR src, VAR tgt, t])) =
    (oput "    ";
     oput src;
     oput " -> ";
     oput tgt;
     oput " : ";
     pp_eva_term t;
     oput "\n")
  | pp_transition t = format_error ("trans", t);

fun pp_transitions l =
    iterate
     pp_transition t
    | t in list l
    end;

fun pp_as (LINE (t, linfo)) =
    (pp_as t handle FailEvt msg => trans_fatal (msg, linfo))
  | pp_as (APP ("as", [t, u])) =
    (pp_eva_term t;
     oput " = ";
     pp_eva_term u)
  | pp_as t = format_error ("as", t);

fun pp_knows (LINE (t, linfo)) =
    (pp_knows t handle FailEvt msg => trans_fatal (msg, linfo))
  | pp_knows (APP ("knows", nil)) = ()
  | pp_knows (APP ("knows", l)) =
    let val sep = ref "  ("
    in
	iterate
	  (oput (!sep);
	   sep := "\n   ";
	   pp_as t)
	| t in list l
	end;
	oput ")\n"
    end
  | pp_knows t = format_error ("knows", t);

fun pp_process (LINE (t, linfo)) =
    (pp_process t handle FailEvt msg => trans_fatal (msg, linfo))
  | pp_process (APP ("process", VAR name::VAR start::knows::l)) =
    (oput "  process ";
     oput name;
     oput " : ";
     oput start;
     oput " {\n";
     pp_transitions l;
     oput "  }\n";
     pp_knows knows;
     oput "\n")
  | pp_process (APP ("repeat-process", (APP ("private", privs))::VAR name::VAR start::knows::l)) =
    (oput "  process*{";
     let val sep = ref ""
     in
	 iterate
	   (oput (!sep);
	    sep := ",";
	    oput x)
	 | VAR x in list privs
	 end
     end;
     oput "} ";
     oput name;
     oput " : ";
     oput start;
     oput " {\n";
     pp_transitions l;
     oput "  }\n";
     pp_knows knows;
     oput "\n")
  | pp_process t = format_error ("process", t);

fun pp_assum (LINE (t, linfo)) =
    (pp_assum t handle FailEvt msg => trans_fatal (msg, linfo))
  | pp_assum t =
    (oput "  assume ";
     pp_eva_term t;
     oput "\n")
  | pp_assum t = format_error ("assume", t);

fun pp_claim (LINE (t, linfo)) =
    (pp_claim t handle FailEvt msg => trans_fatal (msg, linfo))
  | pp_claim t =
    (oput "  claim ";
     pp_eva_term t;
     oput "\n")
  | pp_claim t = format_error ("claim", t);

fun pp_decl (LINE (t, linfo)) =
    (pp_decl t handle FailEvt msg => trans_fatal (msg, linfo))
  | pp_decl (APP ("types", l)) =
    iterate
      pp_type typ
    | typ in list l
    end
  | pp_decl (APP ("values", l)) =
    iterate
      pp_alias t
    | t in list l
    end
  | pp_decl (APP ("axioms", l)) =
    iterate
      pp_axiom t
    | t in list l
    end
  | pp_decl (APP ("system", l)) =
    iterate
      pp_process t
    | t in list l
    end
  | pp_decl (APP ("assume", l)) =
    iterate
      pp_assum t
    | t in list l
    end
  | pp_decl (APP ("claim", l)) =
    iterate
      pp_claim t
    | t in list l
    end
  | pp_decl t = format_error ("decl", t);

fun pp_spec (LINE (t, linfo)) =
    (pp_spec t handle FailEvt msg => trans_fatal (msg, linfo))
  | pp_spec (APP ("compiled-spec", l)) =
    iterate
      pp_decl t
    | t in list l
    end
  | pp_spec t = format_error ("compiled-spec", t);

fun usage () = (#put stderr "Usage: evapp [file]\n";
		#flush stderr ();
		quit 10);

fun main () =
    let val t =
	    case args () of
		[filename] => (let val f as |[close, ...]| = infile filename
			       in
				   lispread f before close ()
			       end handle IO n => (#put stderr "evapp: ";
						   #put stderr filename;
						   #put stderr ": ";
						   #put stderr (iomsg n);
						   #put stderr "\n";
						   #flush stderr ();
						   quit 10))
	      | nil => lispread stdin
	      | _ => usage ()
    in
	pp_spec t;
	oflush ()
    end;
